
import pprint
from datetime import datetime
from rdflib import FOAF, RDF, Literal, URIRef  # type: ignore

import rdflib
from pathlib import Path
from provo.provontologygraph import ProvOntologyGraph

THIS_DIR = Path(__file__).resolve().parent


def test_graph_isomorphism_short_ex():
    graph = ProvOntologyGraph()

    poster = graph.add_entity(
        "https://doi.org/10.6084/m9.figshare.23561661", label="this graphic"
    )
    script = graph.add_activity(
        "https://github.com/rue-a/provo-poster", "provo script")
    me = graph.add_agent("https://orcid.org/0000-0001-8637-9071", "author")
    slub = graph.add_agent("https://ror.org/03wf51b65", "SLUB")

    poster.was_generated_by(script)
    poster.was_attributed_to(me)
    script.was_associated_with(me)
    me.acted_on_behalf_of(slub)

    target_graph = rdflib.Graph()
    target_graph.parse(THIS_DIR / "rdf_output_short_ex.ttl")
    # Assert that the sets of triples are equal
    assert set(graph.get_rdflib_graph()) == set(target_graph)


def test_graph_isomorphism_long_ex():

    prov_ontology_graph = ProvOntologyGraph(
        namespace='http://example.org#',
        namespace_abbreviation=""
    )

    # create entities
    crime_data = prov_ontology_graph.add_entity(
        id='crimeData', label='Crime Data', use_namespace=True)
    national_regions_list = prov_ontology_graph.add_entity(
        id='nationalRegionsList', label='National Regions List', use_namespace=True)
    aggregated_by_regions = prov_ontology_graph.add_entity(
        id='aggregatedByRegions', label='Aggregated by Regions', use_namespace=True)
    bar_chart = prov_ontology_graph.add_entity(
        id='bar_chart', label='Bar Chart', use_namespace=True)

    # create activities
    aggregation_activity = prov_ontology_graph.add_activity(
        id='aggregationActivity', label='Aggregation Activity', use_namespace=True)
    illustration_activity = prov_ontology_graph.add_activity(
        id='illustrationActivity', label='Illustration Activity', use_namespace=True)

    # create agents
    government = prov_ontology_graph.add_agent(
        id='government', label='Government', use_namespace=True)
    civil_action_group = prov_ontology_graph.add_agent(
        id='civil_action_group', label='Civil Action Group', use_namespace=True)
    national_newspaper_inc = prov_ontology_graph.add_agent(
        id='national_newspaper_inc', label='National Newspaper Inc.', use_namespace=True)
    derek = prov_ontology_graph.add_agent(
        id='derek', label='Derek', use_namespace=True)

    # build relations
    crime_data.was_attributed_to(government)
    national_regions_list.was_attributed_to(civil_action_group)

    aggregation_activity.used(crime_data)
    aggregation_activity.used(national_regions_list)
    aggregation_activity.started_at_time(datetime(2011, 7, 14, 1, 1, 1))
    aggregation_activity.ended_at_time(datetime(2011, 7, 14, 2, 2, 2))
    aggregation_activity.was_associated_with(derek)

    aggregated_by_regions.was_generated_by(aggregation_activity)
    aggregated_by_regions.was_attributed_to(derek)

    illustration_activity.was_informed_by(aggregation_activity)
    illustration_activity.used(aggregated_by_regions)
    illustration_activity.was_associated_with(derek)

    bar_chart.was_generated_by(illustration_activity)
    bar_chart.was_derived_from(aggregated_by_regions)
    bar_chart.was_attributed_to(derek)

    derek.acted_on_behalf_of(national_newspaper_inc)

    # print graph to terminal
    print(prov_ontology_graph)

    # use rdflib interface to add FOAF triples
    rdflib_graph = prov_ontology_graph.get_rdflib_graph()

    rdflib_graph.bind("foaf", FOAF)

    rdflib_graph.add((
        URIRef(government.node_id),
        RDF.type,
        FOAF.Organization
    ))

    rdflib_graph.add((
        URIRef(civil_action_group.node_id),
        RDF.type,
        FOAF.Organization
    ))

    rdflib_graph.add((
        URIRef(national_newspaper_inc.node_id),
        RDF.type,
        FOAF.Organization
    ))

    rdflib_graph.add((
        URIRef(national_newspaper_inc.node_id),
        FOAF.name,
        Literal(national_newspaper_inc.label, lang="en")
    ))

    rdflib_graph.add((
        URIRef(derek.node_id),
        RDF.type,
        FOAF.Person
    ))

    rdflib_graph.add((
        URIRef(derek.node_id),
        FOAF.givenName,
        Literal(derek.label, lang="en")
    ))

    rdflib_graph.add((
        URIRef(derek.node_id),
        FOAF.mbox,
        URIRef("mailto:derek@example.org")
    ))

    target_graph = rdflib.Graph()
    target_graph.parse(THIS_DIR / "rdf_output_long_ex.ttl")
    rdflib_graph.serialize(THIS_DIR / 'test.ttl')

    # Assert that the sets of triples are equal
    assert set(rdflib_graph) == set(target_graph)



"""
@author Arne Rümmler
@contact arne.ruemmler@gmail.com

@summary Implementation of a provenance graph
    class that uses PROV-O terms
"""


from dataclasses import dataclass, field

from rdflib import (
    DC,
    FOAF,
    PROV,
    RDF,
    RDFS,
    XSD,
    Graph,  # type: ignore
    Literal,
    Namespace,
    URIRef,
)
from validators import url

from provo.idvault import IdVault
from provo.startingpointclasses import Activity, Agent, Entity
from provo.staticfunctions import update_dict, get_option


@dataclass(frozen=True)
class NamespaceMalformed(Exception):
    """Raised when the namespace is not a valid URL."""

    message: str


@dataclass(frozen=True)
class NamespaceHasNoEndSymbol(Exception):
    """Raised if the namespace does not end with '/' or '#'."""

    message: str


@dataclass(frozen=True)
class PrefixShorthandNotValid(Exception):
    """Raised if the namespace abbreviation contains non-allowed characters."""

    message: str


@dataclass(frozen=True)
class PrefixNotAllowed(Exception):
    """Raised the namespace is not allowed."""

    message: str


@dataclass
class ProvOntologyGraph:
    """model that manages contents of a provenance graph
    that adheres to the PROV-O provenance model."""

    namespace: str = "https://provo-example.org/"
    namespace_abbreviation: str = ""
    lang: str = "en"
    _entities: list[Entity] = field(init=False, default_factory=list)
    _activities: list[Activity] = field(init=False, default_factory=list)
    _agents: list[Agent] = field(init=False, default_factory=list)
    _id_vault: IdVault = field(init=False, default_factory=IdVault)

    def __post_init__(self):
        """check validity of namespace and and namespace abbreviation"""

        # validate namespace
        # TODO rethink validation, we technically want to validate an IRI
        if not url(self.namespace):  # type: ignore
            raise NamespaceMalformed(
                """
            The provided namespace is not a valid URL!

            See validators package (https://github.com/python-validators/validators),
            validators.url() or https://gist.github.com/dperini/729294 for more
            information on what qualifies a URL as valid.
            
            https://www.rfc-editor.org/rfc/rfc3986#section-3:
            3.  Syntax Components

            The generic URI syntax consists of a hierarchical sequence of
            components referred to as the scheme, authority, path, query, and
            fragment.

                URI         = scheme ":" hier-part [ "?" query ] [ "#" fragment ]

                hier-part   = "//" authority path-abempty
                            / path-absolute
                            / path-rootless
                            / path-empty

            The scheme and path components are required, though the path may be
            empty (no characters).  When authority is present, the path must
            either be empty or begin with a slash ("/") character.  When
            authority is not present, the path cannot begin with two slash
            characters ("//").  These restrictions result in five different ABNF
            rules for a path (Section 3.3), only one of which will match any
            given URI reference.

            The following are two example URIs and their component parts:

                    foo://example.com:8042/over/there?name=ferret#nose
                    \_/   \______________/\_________/ \_________/ \__/
                    |           |            |            |        |
                scheme     authority       path        query   fragment
                    |   _____________________|__
                    / \ /                        \
                    urn:example:animal:ferret:nose
            """
            )
        end_symbol = self.namespace[-1]
        if end_symbol not in ("/", "#"):
            raise NamespaceHasNoEndSymbol(
                "The provided Namespace has to end on a '/' or '#'!"
            )

        # validate namespace abbreviation
        # TODO check what is actually allowed
        allowed_symbols_for_prefix_shorthand = "abcdefghijklmnopqrstuvwxyz"

        for symbol in self.namespace_abbreviation:
            if symbol not in allowed_symbols_for_prefix_shorthand:
                raise PrefixShorthandNotValid(
                    f"""
                    The provided namespace abbreviation is not valid.
                    
                    Character of the namespace have to be in "{allowed_symbols_for_prefix_shorthand}".
                    """
                )
        # if self.namespace_abbreviation in forbidden_namespaces:
        core_prefixes = ["owl", "rdf", "rdfs", "xsd", "xml"]
        if self.namespace_abbreviation in core_prefixes:
            raise PrefixNotAllowed(
                f"""
            The provided namespace abbreviation is a core prefix, 
            and thus, prohibited from use.

            Core prefixes are: {"".join([f"{k}, " if i != len(core_prefixes)-2 else f"{k} and " for i, k in enumerate(core_prefixes)])[:-2]}.
            """
            )

    def __str__(self) -> str:
        """prints the contents of the provenance graph in a nice format."""
        contents = "Provenance Graph Contents:"
        contents += "\n    ---"
        contents += "\n    Entities:"
        contents += "\n        ---\n"
        for entity in self._entities:
            lines = entity.__str__().splitlines()
            for line in lines:
                contents += f"        {line}\n"
        contents += "    ---"
        contents += "\n    Activities:"
        contents += "\n        ---\n"
        for activity in self._activities:
            lines = activity.__str__().splitlines()
            for line in lines:
                contents += f"        {line}\n"
        contents += "    ---"
        contents += "\n    Agents:"
        contents += "\n        ---\n"
        for agent in self._agents:
            lines = agent.__str__().splitlines()
            for line in lines:
                contents += f"        {line}\n"
        contents += "    ---"

        return contents

    def _handle_id(self, id: str = "", use_namespace: bool = False) -> str:
        """checks whether the provided namespace-id combination is
        already used for a node in the graph.
        if no namespace is provided: default namespace is used,
        if no id is provided: id get automatically generated."""

        if use_namespace:
            id = self.namespace + id
        if id:
            node_id = self._id_vault.add_id(id)
        else:
            node_id = self._id_vault.generate(self.namespace)
        return node_id

    def add_entity(
        self,
        id: str = "",
        label: str = "",
        description: str = "",
        use_namespace: bool = False,
    ) -> Entity:
        """creates a new entity, adds it to the graph and returns it then"""

        node_id = self._handle_id(id, use_namespace)
        entity = Entity(node_id=node_id, label=label, description=description)
        self._entities.append(entity)
        return entity

    def add_activity(
        self,
        id: str = "",
        label: str = "",
        description: str = "",
        use_namespace: bool = False,
    ) -> Activity:
        """creates a new activity, adds it to the graph and returns it then"""

        node_id = self._handle_id(id, use_namespace)
        activity = Activity(node_id=node_id, label=label,
                            description=description)
        self._activities.append(activity)
        return activity

    def add_agent(
        self,
        id: str = "",
        label: str = "",
        description: str = "",
        use_namespace: bool = False,
    ) -> Agent:
        """creates a new agent, adds it to the graph and returns it then"""

        node_id = self._handle_id(id, use_namespace)
        agent = Agent(node_id=node_id, label=label, description=description)
        self._agents.append(agent)
        return agent

    # TODO
    # def read_graph(self, file_path: str) -> None:
    #     """reads the contents of an existing PROV-O provenance graph."""

    #     pass

    def get_rdflib_graph(self) -> Graph:
        """returns the provenance graph as rdflib.Graph()."""

        provenance_graph = Graph()
        provenance_graph.bind("dc", DC)
        provenance_graph.bind("foaf", FOAF)
        provenance_graph.bind("rdf", RDF)
        provenance_graph.bind("rdfs", RDFS)
        provenance_graph.bind("prov", PROV)
        provenance_graph.bind(self.namespace_abbreviation,
                              Namespace(self.namespace))

        for node in self._entities + self._activities + self._agents:
            if node.label:
                provenance_graph.add(
                    (
                        URIRef(node.node_id),
                        RDFS.label,
                        Literal(node.label, lang=self.lang),
                    )
                )
            if node.description:
                provenance_graph.add(
                    (
                        URIRef(node.node_id),
                        RDFS.comment,
                        Literal(node.description, lang=self.lang),
                    )
                )
        for entity in self._entities:
            provenance_graph.add(
                (URIRef(entity.node_id), RDF.type, PROV.Entity))
            for activity in entity._was_generated_by_activities:
                provenance_graph.add(
                    (
                        URIRef(entity.node_id),
                        PROV.wasGeneratedBy,
                        URIRef(activity.node_id),
                    )
                )
            for origin_entity in entity._was_derived_from_entities:
                provenance_graph.add(
                    (
                        URIRef(entity.node_id),
                        PROV.wasDerivedFrom,
                        URIRef(origin_entity.node_id),
                    )
                )
            for agent in entity._was_attributed_to_agents:
                provenance_graph.add(
                    (
                        URIRef(entity.node_id),
                        PROV.wasAttributedTo,
                        URIRef(agent.node_id),
                    )
                )
        for activity in self._activities:
            provenance_graph.add(
                (URIRef(activity.node_id), RDF.type, PROV.Activity))
            if activity._start_time:
                provenance_graph.add(
                    (
                        URIRef(activity.node_id),
                        PROV.startedAtTime,
                        Literal(activity._start_time, datatype=XSD.dateTime),
                    )
                )
            if activity._end_time:
                provenance_graph.add(
                    (
                        URIRef(activity.node_id),
                        PROV.endedAtTime,
                        Literal(activity._end_time, datatype=XSD.dateTime),
                    )
                )
            for entity in activity._used_entities:
                provenance_graph.add(
                    (URIRef(activity.node_id), PROV.used, URIRef(entity.node_id))
                )
            for previous_activity in activity._was_informed_by_activities:
                provenance_graph.add(
                    (
                        URIRef(activity.node_id),
                        PROV.wasInformedBy,
                        URIRef(previous_activity.node_id),
                    )
                )
            for agent in activity._was_associated_with_agents:
                provenance_graph.add(
                    (
                        URIRef(agent.node_id),
                        PROV.wasAssociatedWith,
                        URIRef(agent.node_id),
                    )
                )
        for agent in self._agents:
            provenance_graph.add((URIRef(agent.node_id), RDF.type, PROV.Agent))
            for instructor in agent._acted_on_behalf_of_agents:
                provenance_graph.add(
                    (
                        URIRef(agent.node_id),
                        PROV.actedOnBehalfOf,
                        URIRef(instructor.node_id),
                    )
                )

        return provenance_graph

    def serialize_as_rdf(self, file_name: str, export_format: str = "turtle") -> None:
        """serializes the graph as rdf, available formats are:
        "xml", "n3", "turtle", "nt", "pretty-xml", "trix", "trig", "nquads", "json-ld", "hext"
        """
        self.get_rdflib_graph().serialize(destination=file_name, format=export_format)

    def export_as_mermaid_flowchart(
        self, file_name: str, user_options: dict = {}
    ) -> None:
        """exports the contents of the graph as mermaid-md flowchart"""

        # if possible the options use the mermaid terminology
        default_options = {
            "invert-relations": False,
            "orientation": "TD",
            "included-relations": [
                "was_generated_by",
                "was_attributed_to",
                "used",
                "was_associated_with",
                "acted_on_behalf_of",
            ],
            "color": "#000000",
            "stroke": "#a4a4a4",
            "stroke-width": "1px",
            "relation-style": "-",
            "entity": {
                "fill": "#fffedf",
                "shape": "([:])",
                "relation-style": None,
                "color": None,
                "stroke": None,
                "stroke-width": None,
            },
            "activity": {
                "fill": "#cfceff",
                "shape": "[[:]]",
                "relation-style": None,
                "color": None,
                "stroke": None,
                "stroke-width": None,
            },
            "agent": {
                "fill": "#ffebc3",
                "shape": "[/:\\]",
                "relation-style": ".",
                "color": None,
                "stroke": None,
                "stroke-width": None,
            },
        }

         # Update default options with user options
        options = update_dict(default_options, user_options)

        # Unpack commonly used option values
        text_color = options["color"]
        stroke_color = options["stroke"]
        stroke_width = options["stroke-width"]
        stroke_style = options["relation-style"]

        # Unpack entity options
        entity_opts = options["entity"]
        entity_fill_color = entity_opts["fill"]
        entity_text_color = entity_opts["color"] or text_color
        entity_stroke_color = entity_opts["stroke"] or stroke_color
        entity_stroke_width = entity_opts["stroke-width"] or stroke_width
        entity_stroke_style = entity_opts["relation-style"] or stroke_style

        # Unpack activity options
        activity_opts = options["activity"]
        activity_fill_color = activity_opts["fill"]
        activity_text_color = activity_opts["color"] or text_color
        activity_stroke_color = activity_opts["stroke"] or stroke_color
        activity_stroke_width = activity_opts["stroke-width"] or stroke_width
        activity_stroke_style = activity_opts["relation-style"] or stroke_style

        # Unpack agent options
        agent_opts = options["agent"]
        agent_fill_color = agent_opts["fill"]
        agent_text_color = agent_opts["color"] or text_color
        agent_stroke_color = agent_opts["stroke"] or stroke_color
        agent_stroke_width = agent_opts["stroke-width"] or stroke_width
        agent_stroke_style = agent_opts["relation-style"] or stroke_style



        lines = []
        lines.append("```mermaid")
        lines.append(f"flowchart {options['orientation']}")

        # Class definitions for entities
        lines.append(f"classDef entity fill:{entity_fill_color}")
        if entity_text_color:
            lines.append(f"classDef entity color:{entity_text_color}")
        if entity_stroke_color:
            lines.append(f"classDef entity stroke:{entity_stroke_color}")
        if entity_stroke_width:
            lines.append(f"classDef entity stroke-width:{entity_stroke_width}")

        # Class definitions for activities
        lines.append(f"classDef activity fill:{activity_fill_color}")
        if activity_text_color:
            lines.append(f"classDef activity color:{activity_text_color}")
        if activity_stroke_color:
            lines.append(f"classDef activity stroke:{activity_stroke_color}")
        if activity_stroke_width:
            lines.append(f"classDef activity stroke-width:{activity_stroke_width}")

        # Class definitions for agents
        lines.append(f"classDef agent fill:{agent_fill_color}")
        if agent_text_color:
            lines.append(f"classDef agent color:{agent_text_color}")
        if agent_stroke_color:
            lines.append(f"classDef agent stroke:{agent_stroke_color}")
        if agent_stroke_width:
            lines.append(f"classDef agent stroke-width:{agent_stroke_width}")

        for entity in self._entities:
            lines.append(
                f'{entity.node_id}{options["entity"]["shape"].split(":")[0]}<a style=color:inherit href={entity.node_id}>{entity.label or entity.node_id}</a>{options["entity"]["shape"].split(":")[1]}:::entity'
            )
            if "was_derived_by" in options["included-relations"]:
                attrs = "style=color:inherit href=https://www.w3.org/TR/prov-o/#wasDerivedFrom"
                for item in entity._was_derived_from_entities:                    
                    if not options["invert-relations"]:
                        lines.append(
                            f"{entity.node_id}-{entity_stroke_style} <a {attrs}>was derived from</a> {entity_stroke_style}->{item.node_id}"
                        )
                    else:
                        lines.append(
                            f"{item.node_id}-{entity_stroke_style} <a {attrs}>originated from</a> {entity_stroke_style}->{entity.node_id}"
                        )
            if "was_generated_by" in options["included-relations"]:
                attrs ="style=color:inherit href=https://www.w3.org/TR/prov-o/#wasGeneratedBy"
                for item in entity._was_generated_by_activities:
                    
                    if not options["invert-relations"]:
                        lines.append(
                            f"{entity.node_id}-{activity_stroke_style} <a {attrs}>was generated by</a> {activity_stroke_style}->{item.node_id}"
                        )
                    else:
                        lines.append(
                            f"{item.node_id}-{activity_stroke_style} <a {attrs}>generated</a> {activity_stroke_style}->{entity.node_id}"
                        )
            if "was_attributed_to" in options["included-relations"]:
                attrs = "style=color:inherit href=https://www.w3.org/TR/prov-o/#wasAttributedTo"
                for item in entity._was_attributed_to_agents:
                    
                    if not options["invert-relations"]:
                        lines.append(
                            f"{entity.node_id}-{agent_stroke_style} <a {attrs}>was attributed to</a> {agent_stroke_style}->{item.node_id}"
                        )
                    else:
                        lines.append(
                            f"{item.node_id}-{agent_stroke_style} <a {attrs}>produced</a> {agent_stroke_style}->{entity.node_id}"
                        )

        for activity in self._activities:
            lines.append(
                f"{activity.node_id}{options['activity']['shape'].split(':')[0]}<a style=color:inherit href={activity.node_id}>{activity.label or activity.node_id}</a>{options['activity']['shape'].split(':')[1]}:::activity"
            )
            if "was_informed_by" in options["included-relations"]:
                attrs = "style=color:inherit href=https://www.w3.org/TR/prov-o/#wasInformedBy"
                for item in activity._was_informed_by_activities:
                    
                    if not options["invert-relations"]:
                        lines.append(
                            f"{activity.node_id}-{entity_stroke_style} <a {attrs}>was informed by</a> {entity_stroke_style}->{item.node_id}"
                        )
                    else:
                        lines.append(
                            f"{item.node_id}-{entity_stroke_style} <a {attrs}>informed</a> {entity_stroke_style}->{activity.node_id}"
                        )
            if "used" in options["included-relations"]:
                attrs="style=color:inherit href=https://www.w3.org/TR/prov-o/#used"
                for item in activity._used_entities:                    
                    if not options["invert-relations"]:
                        lines.append(
                            f"{activity.node_id}-{activity_stroke_style} <a {attrs}>used</a> {activity_stroke_style}->{item.node_id}"
                        )
                    else:
                        lines.append(
                            f"{item.node_id}-{activity_stroke_style} <a {attrs}>was used by</a> {activity_stroke_style}->{activity.node_id}"
                        )
            if "was_associated_with" in options["included-relations"]:
                attrs = "style=color:inherit href=https://www.w3.org/TR/prov-o/#wasAssociatedWith"
                for item in activity._was_associated_with_agents:
                    if not options["invert-relations"]:
                        lines.append(
                            f"{activity.node_id}-{agent_stroke_style} <a {attrs}>was associated with</a> {agent_stroke_style}->{item.node_id}"
                        )
                    else:
                        lines.append(
                            f"{item.node_id}-{agent_stroke_style} <a {attrs}>initiated</a> {agent_stroke_style}->{activity.node_id}"
                        )

        for agent in self._agents:
            lines.append(
                f"{agent.node_id}{options['agent']['shape'].split(':')[0]}<a style=color:inherit href={agent.node_id}>{agent.label or agent.node_id}</a>{options['agent']['shape'].split(':')[1]}:::agent"
            )
            if "acted_on_behalf_of" in options["included-relations"]:
                attrs = "style=color:inherit href=https://www.w3.org/TR/prov-o/#actedOnBehalfOf"
                for item in agent._acted_on_behalf_of_agents:
                    if not options["invert-relations"]:
                        lines.append(
                            f"{agent.node_id}-{agent_stroke_style} <a {attrs}>acted on behalf of</a> {agent_stroke_style}->{item.node_id}"
                        )
                    else:
                        lines.append(
                            f"{item.node_id}-{agent_stroke_style} <a {attrs}>instructed</a> {agent_stroke_style}->{agent.node_id}"
                        )

        lines.append("```")
        lines = "\n".join(lines)

        with open(file_name, "w") as f:
            f.write(lines)



"""
@author Arne Rümmler
@contact arne.ruemmler@gmail.com

@summary Implementation of the starting term classes of
 PROV-O https://www.w3.org/TR/prov-o/ .
"""


from dataclasses import dataclass, field
from uuid import uuid4
from validators import url


@dataclass(frozen=True)
class IdMalformed(Exception):
    """Raised when a node id is invalid."""

    message: str


@dataclass(frozen=True)
class IdAlreadyUsed(Exception):
    """Raised when a node with an id that already exists is created."""

    message: str


@dataclass
class IdVault:
    """manages unique IDs and generates new ones"""

    vault: list[str] = field(init=False, default_factory=list)
    _invalid_symbols: str = field(init=False, default='<>" {}|\\^`')

    def _raise_exception_if_uri_invalid(self, uri: str) -> None:
        "checks if uri is valid and raises exception if not"

        # TODO rethink validation, we technically want to validate an IRI
        if not url(uri):  # type: ignore
            raise IdMalformed(
                """
            The provided namespace is not a valid URL!

            See validators package (https://github.com/python-validators/validators),
            validators.url() or https://gist.github.com/dperini/729294 for more
            information on what qualifies a URL as valid.
            
            https://www.rfc-editor.org/rfc/rfc3986#section-3:
            3.  Syntax Components

            The generic URI syntax consists of a hierarchical sequence of
            components referred to as the scheme, authority, path, query, and
            fragment.

                URI         = scheme ":" hier-part [ "?" query ] [ "#" fragment ]

                hier-part   = "//" authority path-abempty
                            / path-absolute
                            / path-rootless
                            / path-empty

            The scheme and path components are required, though the path may be
            empty (no characters).  When authority is present, the path must
            either be empty or begin with a slash ("/") character.  When
            authority is not present, the path cannot begin with two slash
            characters ("//").  These restrictions result in five different ABNF
            rules for a path (Section 3.3), only one of which will match any
            given URI reference.

            The following are two example URIs and their component parts:

                    foo://example.com:8042/over/there?name=ferret#nose
                    \_/   \______________/\_________/ \_________/ \__/
                    |           |            |            |        |
                scheme     authority       path        query   fragment
                    |   _____________________|__
                    / \ /                        \
                    urn:example:animal:ferret:nose
            """
            )

    def generate(self, namespace: str) -> str:
        """generates a new uuid that is not in the vault yet"""
        node_id = namespace + str(uuid4())
        while node_id in self.vault:
            node_id = namespace + str(uuid4())
        self._raise_exception_if_uri_invalid(node_id)
        self.vault.append(node_id)
        return node_id

    def add_id(self, node_id: str) -> str:
        """adds a user defined id and checks if its valid"""
        if node_id in self.vault:
            raise IdAlreadyUsed(
                f'The Id "{node_id}" was already used in this graph.')
        print(node_id)
        self._raise_exception_if_uri_invalid(node_id)
        self.vault.append(node_id)
        return node_id



"""
@author Arne Rümmler
@contact arne.ruemmler@gmail.com

@summary Implementation of the starting term classes of
 PROV-O https://www.w3.org/TR/prov-o/ .
"""


from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime


@dataclass(frozen=True)
class NoStartTimeDefined(Exception):
    """Raised if user user wants to access the start time of
    an activity for which no start time was defined."""

    message: str


@dataclass(frozen=True)
class NoEndTimeDefined(Exception):
    """Raised if user user wants to access the end time of
    an activity for which no end time was defined."""

    message: str


@dataclass(frozen=True)
class InvalidProvClassForThisRelation(Exception):
    """Raised when the wrong class is given as attribute
    to one of the starting point property methods."""

    message: str


@dataclass(frozen=True)
class NoDateTime(Exception):
    """Raised when no datetime object is given as attribute."""

    message: str


@dataclass
class Node(ABC):
    """ Abstract parent class of Activity, Agent, and Entity. """

    label: str = ""
    description: str = ""
    node_id: str = ""

    @abstractmethod
    def __str__(self) -> str:
        """prints the node in a nice format"""
        contents = f"id: {self.node_id}"
        if self.label:
            contents += f"\nlabel: {self.label}"
        if self.description:
            contents += f"\ndescription: {self.description}"
        return contents


@dataclass
class Entity(Node):
    """ Class to create a PROV-O Entity object. 
    https://www.w3.org/TR/prov-o/#Entity """

    _was_derived_from_entities: list['Entity'] = field(
        init=False, default_factory=list)
    _was_generated_by_activities: list['Activity'] = field(
        init=False, default_factory=list)
    _was_attributed_to_agents: list['Agent'] = field(
        init=False, default_factory=list)

    def __str__(self) -> str:
        """prints the entity in a nice format."""
        contents = super().__str__()
        if self._was_derived_from_entities:
            contents += f"\nwas derived from: {[item.label if item.label else item.node_id for item in self._was_derived_from_entities]}"
        if self._was_generated_by_activities:
            contents += f"\nwas generated by: {[item.label if item.label else item.node_id for item in self._was_generated_by_activities]}"
        if self._was_attributed_to_agents:
            contents += f"\nwas attributed to: {[item.label if item.label else item.node_id for item in self._was_attributed_to_agents]}"
        contents += "\n---"
        return contents

    def was_derived_from(self, entity: 'Entity') -> None:
        """ implements the wasDerivedFrom property of PROV-O
        https://www.w3.org/TR/prov-o/#wasDerivedFrom """

        if not isinstance(entity, Entity):
            raise InvalidProvClassForThisRelation(
                f"""The PROV relation "was derived from" refers to an Entity, thus the provided
                attribute has to be of type <class 'provo.startingpointclasses.Entity'>. 
                In contradiction to this, the provided attribute is of the type {type(entity)}."""
            )
        self._was_derived_from_entities.append(entity)

    def was_generated_by(self, activity: 'Activity') -> None:
        """ implements the wasGeneratedBy property of PROV-O
        https://www.w3.org/TR/prov-o/#wasgeneratedBy """

        if not isinstance(activity, Activity):
            raise InvalidProvClassForThisRelation(
                f"""The PROV relation "was generated by" refers to an Activity, thus the provided
                attribute has to be of type <class 'provo.startingpointclasses.Activity'>. 
                In contradiction to this, the provided attribute is of the type {type(activity)}."""
            )
        self._was_generated_by_activities.append(activity)

    def was_attributed_to(self, agent: 'Agent') -> None:
        """ implements the wasAttributedTo property of PROV-O
        https://www.w3.org/TR/prov-o/#wasAttributedTo """

        if not isinstance(agent, Agent):
            raise InvalidProvClassForThisRelation(
                f"""The PROV relation "was attributed to" refers to an Agent, thus the provided
                attribute has to be of type <class 'provo.startingpointclasses.Agent'>. 
                In contradiction to this, the provided attribute is of the type {type(agent)}."""
            )
        self._was_attributed_to_agents.append(agent)


@dataclass
class Activity(Node):
    """ Class to create a PROV-O Activity object.
    https://www.w3.org/TR/prov-o/#Activity """

    _was_informed_by_activities: list['Activity'] = field(
        init=False, default_factory=list)
    _used_entities: list['Entity'] = field(init=False, default_factory=list)
    _was_associated_with_agents: list['Agent'] = field(
        init=False, default_factory=list)
    # type: ignore (there is no default date that resolves to False)
    _start_time: datetime = field(init=False, default_factory=lambda: None)
    _end_time: datetime = field(
        init=False, default_factory=lambda: None)  # type: ignore

    def __str__(self) -> str:
        """prints the activity in a nice format."""
        contents = super().__str__()
        if self._start_time:
            contents += f"\nstart time: {self._start_time}"
        if self._end_time:
            contents += f"\nend time: {self._end_time}"
        if self._was_informed_by_activities:
            contents += f"\nwas informed by: {[item.label if item.label else item.node_id for item in self._was_informed_by_activities]}"
        if self._used_entities:
            contents += f"\nused: {[item.label if item.label else item.node_id for item in self._used_entities]}"
        if self._was_associated_with_agents:
            contents += f"\nwas associated with: {[item.label if item.label else item.node_id for item in self._was_associated_with_agents]}"
        contents += "\n---"
        return contents

    def get_start_time(self) -> datetime:
        """returns the start time of the activity"""
        if self._start_time:
            return self._start_time
        else:
            raise NoStartTimeDefined(
                f"No start time defined for the activity {self.label if self.label else f'with the id <{self.node_id}>'}.")

    def get_end_time(self):
        """returns the end time of the activity"""
        if self._end_time:
            return self._end_time
        else:
            raise NoEndTimeDefined(
                f"No end time defined for the activity {self.label if self.label else f'with the id <{self.node_id}>'}.")

    def was_informed_by(self, activity: 'Activity') -> None:
        """ implements the wasInformedBy property of PROV-O
        https://www.w3.org/TR/prov-o/#wasInformedBy """

        if not isinstance(activity, Activity):
            raise InvalidProvClassForThisRelation(
                f"""The PROV relation "was informed by" refers to an Activity, thus the provided
                attribute has to be of type <class 'provo.startingpointclasses.Activity'>. 
                In contradiction to this, the provided attribute is of the type {type(activity)}."""
            )
        self._was_informed_by_activities.append(activity)

    def used(self, entity: 'Entity') -> None:
        """ implements the used property of PROV-O
        https://www.w3.org/TR/prov-o/#used """

        if not isinstance(entity, Entity):
            raise InvalidProvClassForThisRelation(
                f"""The PROV relation "used" refers to an Entity, thus the provided
                attribute has to be of type <class 'provo.startingpointclasses.Entity'>. 
                In contradiction to this, the provided attribute is of the type {type(entity)}."""
            )
        self._used_entities.append(entity)

    def was_associated_with(self, agent: 'Agent') -> None:
        """ implements the wasAssociatedWith property of PROV-O
        https://www.w3.org/TR/prov-o/#wasAssociatedWith """

        if not isinstance(agent, Agent):
            raise InvalidProvClassForThisRelation(
                f"""The PROV relation "was associated with" refers to an Agent, thus the provided
                attribute has to be of type <class 'provo.startingpointclasses.Agent'>. 
                In contradiction to this, the provided attribute is of the type {type(agent)}."""
            )
        self._was_associated_with_agents.append(agent)

    def started_at_time(self, start_time: datetime) -> None:
        """ implements the startedAtTime property of PROV-O
        https://www.w3.org/TR/prov-o/#startedAtTime """

        if not isinstance(start_time, datetime):
            raise NoDateTime(
                "The attribute `start_time` of the method `started_at_time` has to be a `datetime` object.")
        self._start_time = start_time

    def ended_at_time(self, end_time: datetime) -> None:
        """ implements the endedAtTime property of PROV-O
        https://www.w3.org/TR/prov-o/#endedAtTime """

        if not isinstance(end_time, datetime):
            raise NoDateTime(
                "The attribute `end_time` of the method `ended_at_time` has to be a `datetime` object.")

        self._end_time = end_time


@dataclass
class Agent(Node):
    """ Class to create a PROV-O Agent object.
    https://www.w3.org/TR/prov-o/#Agent """

    _acted_on_behalf_of_agents: list['Agent'] = field(
        init=False, default_factory=list)

    def __str__(self) -> str:
        """prints the agent in a nice format."""
        contents = super().__str__()
        if self._acted_on_behalf_of_agents:
            contents += f"\nacted on behalf of: {[item.label if item.label else item.node_id for item in self._acted_on_behalf_of_agents]}"
        contents += "\n---"
        return contents

    def acted_on_behalf_of(self, agent: 'Agent') -> None:
        """ implements the actedOnBehalfOf property of PROV-O
        https://www.w3.org/TR/prov-o/#actedOnBehalfOf """

        if not isinstance(agent, Agent):
            raise InvalidProvClassForThisRelation(
                f"""The PROV relation "acted on behalf of" refers to an Agent, thus the provided
                attribute has to be of type <class 'provo.startingpointclasses.Agent'>. 
                In contradiction to this, the provided attribute is of the type {type(agent)}."""
            )
        self._acted_on_behalf_of_agents.append(agent)



# from provo.startingpointclasses import Activity, Agent, Entity


def update_dict(dict_old: dict, dict_with_updates: dict) -> dict:
    """updates/adds the fields from dict_with_updates in/to dict_old"""

    for key, value in dict_with_updates.items():
        try:
            value = dict(value)
            value = update_dict(dict_old[key], dict_with_updates=value)
        except:
            dict_old[key] = value
    return dict_old


def get_option(option, sub_option, default_option):
    return option[sub_option] if option.get(sub_option) else default_option



# -*- coding: utf-8 -*-
"""
provo
-----

Construct  PROV-O compliant provenance graphs.
"""

from pathlib import Path

from setuptools import find_packages, setup

this_directory = Path(__file__).parent

setup(
    name="provo",
    author="Arne Rümmler",
    author_email="arne.ruemmler@gmail.com",
    version="1.1.1",
    description="Construct  PROV-O compliant provenance graphs.",
    url="https://github.com/rue-a/provo",
    packages=find_packages(".", exclude=["tests", "tests.*"]),
    install_requires=["rdflib", "validators", "uuid"],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    python_requires=">=3.09",
)



import pytest
from provo.idvault import IdAlreadyUsed, IdMalformed, IdVault
from validators import url


def test_generate():
    """tests if the automatically generated id is a valid url"""

    namespace = "https://test.package/"
    vault = IdVault()
    id = vault.generate(namespace=namespace)
    # TODO rethink validation
    assert url(id)  # type: ignore


def test_no_id_duplication():
    """tests if vault throws IdAlreadyUsed exception when an existing
    id is added to the vault"""

    test_id = "https://test.package/test"
    vault = IdVault()
    vault.add_id(node_id=test_id)
    with pytest.raises(IdAlreadyUsed):
        vault.add_id(node_id=test_id)


def test_id_validation():
    """tests if vault throws IdMalformed exception if id contains one
    of the invalid symbols"""

    vault = IdVault()
    invalid_symbols = '<>" {}|\\^`'

    for symbol in invalid_symbols:
        # test if error is thrown if namespace contains the symbol
        namespace = f"https://t{symbol}est.package/"
        test_id = "test"
        with pytest.raises(IdMalformed):
            vault._raise_exception_if_uri_invalid(namespace + test_id)

        # test if error is thrown if id without namespace contains the symbol
        namespace = "https://test.package/"
        test_id = f"te{symbol}st"
        with pytest.raises(IdMalformed):
            vault._raise_exception_if_uri_invalid(namespace + test_id)




import pytest

from provo.provontologygraph import (NamespaceHasNoEndSymbol,
                                     NamespaceMalformed, PrefixNotAllowed,
                                     PrefixShorthandNotValid,
                                     ProvOntologyGraph)


def test_graph_initialization():
    """tests if namespace and namespace abbreviation (i.e., prefix) is valid"""
    
    ProvOntologyGraph()
    ProvOntologyGraph(namespace="https://test.package/", namespace_abbreviation="")
    ProvOntologyGraph(namespace="https://test.package#", namespace_abbreviation="")

    malformed_namespaces = ["www.test.de", "https://test.package/ ",
                            "test.org", " https://test.package/"]
    for namespace in malformed_namespaces:
        with pytest.raises(NamespaceMalformed):
            ProvOntologyGraph(namespace=namespace)

    with pytest.raises(NamespaceHasNoEndSymbol):
        ProvOntologyGraph(namespace="https://test.package")

    invalid_prefixes = [" ", ":", "test:"]
    for prefix in invalid_prefixes:
        with pytest.raises(PrefixShorthandNotValid):
            ProvOntologyGraph(namespace_abbreviation=prefix)

    core_prefixes = ["owl", "rdf", "rdfs", "xsd", "xml"]
    for prefix in core_prefixes:
        with pytest.raises(PrefixNotAllowed):
            ProvOntologyGraph(namespace_abbreviation=prefix)



from datetime import datetime

import pytest
from provo.startingpointclasses import (Activity, Agent, Entity,
                                        InvalidProvClassForThisRelation,
                                        NoDateTime)

entity = Entity(node_id='https://test.package/entity')
activity = Activity(node_id='https://test.package/activity')
agent = Agent(node_id='https://test.package/agent')


def test_was_derived_from():
    """tests was derived from method"""
    entity.was_derived_from(entity)
    with pytest.raises(InvalidProvClassForThisRelation):
        entity.was_derived_from(activity)  # type: ignore
    with pytest.raises(InvalidProvClassForThisRelation):
        entity.was_derived_from(agent)  # type: ignore


def test_was_generated_by():
    """tests was generated by method"""
    entity.was_generated_by(activity)
    with pytest.raises(InvalidProvClassForThisRelation):
        entity.was_generated_by(entity)  # type: ignore
    with pytest.raises(InvalidProvClassForThisRelation):
        entity.was_generated_by(agent)  # type: ignore


def test_was_attributed_to():
    """tests was attributed to method"""
    entity.was_attributed_to(agent)
    with pytest.raises(InvalidProvClassForThisRelation):
        entity.was_attributed_to(activity)  # type: ignore
    with pytest.raises(InvalidProvClassForThisRelation):
        entity.was_attributed_to(entity)  # type: ignore


def test_was_informed_by():
    """tests was informed by method"""
    activity.was_informed_by(activity)
    with pytest.raises(InvalidProvClassForThisRelation):
        activity.was_informed_by(entity)  # type: ignore
    with pytest.raises(InvalidProvClassForThisRelation):
        activity.was_informed_by(agent)  # type: ignore


def test_was_associated_with():
    """tests was associated with method"""
    activity.was_associated_with(agent)
    with pytest.raises(InvalidProvClassForThisRelation):
        activity.was_associated_with(activity)  # type: ignore
    with pytest.raises(InvalidProvClassForThisRelation):
        activity.was_associated_with(entity)  # type: ignore


def test_used():
    """tests used method"""
    activity.used(entity)
    with pytest.raises(InvalidProvClassForThisRelation):
        activity.used(activity)  # type: ignore
    with pytest.raises(InvalidProvClassForThisRelation):
        activity.used(agent)  # type: ignore


def test_acted_on_behalf_of():
    """tests acted on behalf of method"""
    agent.acted_on_behalf_of(agent)
    with pytest.raises(InvalidProvClassForThisRelation):
        agent.acted_on_behalf_of(activity)  # type: ignore
    with pytest.raises(InvalidProvClassForThisRelation):
        agent.acted_on_behalf_of(entity)  # type: ignore


def test_started_at_time():
    """tests started at time method"""
    activity.started_at_time(datetime(1980, 1, 1))
    with pytest.raises(NoDateTime):
        activity.started_at_time("01.01.1980")  # type: ignore
    with pytest.raises(NoDateTime):
        activity.started_at_time(1980)  # type: ignore


def test_ended_at_time():
    """tests ended at time method"""
    activity.ended_at_time(datetime(1981, 1, 1))
    with pytest.raises(NoDateTime):
        activity.ended_at_time("01.01.1981")  # type: ignore
    with pytest.raises(NoDateTime):
        activity.ended_at_time(1981)  # type: ignore
